#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <conio.h>

#include "controller.h"
#include "mill.h"

static void JogScreen(void);

static HANDLE Serial;

static BOOL LastXWasNegative;

#define Write(b, len) do { \
        DWORD written; \
        WriteFile(Serial, b, len, &written, NULL); \
    } while(0)

//-----------------------------------------------------------------------------
// Read from the controller until we have a \r\n line ending.
//-----------------------------------------------------------------------------
static GetLine(BYTE *dest)
{
    int p = 0;
    BYTE b;
    DWORD r;

    do {
        if(!ReadFile(Serial, &b, 1, &r, NULL)) {
            printf("read error\n");
            exit(-1);
        }
        if(r == 1) {
            dest[p++] = b;
        }
    } while(p < 2 || dest[p-1] != '\n' || dest[p-2] != '\r');
}

//-----------------------------------------------------------------------------
// Use our fastest speed (for positioning only; will trash cutters!)
//-----------------------------------------------------------------------------
static void Fast(void)
{
    BYTE setDefPeriod[2] = { SET_DEF_PERIOD, 40 };
    BYTE resp[5];
    Write(setDefPeriod, 2);
    GetLine(resp);
}

//-----------------------------------------------------------------------------
// Use our slower speed (for drilling, milling).
//-----------------------------------------------------------------------------
static void Slow(void)
{
    BYTE setDefPeriod[2] = { SET_DEF_PERIOD, 80 };
    BYTE resp[5];
    Write(setDefPeriod, 2);
    GetLine(resp);
}

void mill_translate(int dx, int dy)
{
    DWORD written;
    BYTE buf[6];
    BYTE resp[5];

    if(dx != 0) {
        if(LastXWasNegative && dx > 0) {
            buf[0] = TRANSLATE;
            buf[1] = 0;
            buf[2] = 0;
            buf[3] = 16;
            buf[4] = 0;
            buf[5] = 0;
            Write(buf, 6);
            GetLine(resp);
        } else if(!LastXWasNegative && dx < 0) {
            buf[0] = TRANSLATE;
            buf[1] = 1;
            buf[2] = 0;
            buf[3] = 16;
            buf[4] = 0;
            buf[5] = 0;
            Write(buf, 6);
            GetLine(resp);
        }

        if(dx < 0)
            LastXWasNegative = TRUE;
        else
            LastXWasNegative = FALSE;
    }

    buf[0] = TRANSLATE;
    buf[1] = ((dx > 0) ? 0 : 1) | ((dy > 0) ? 0 : 2);
    buf[2] = abs(dx) >> 8;
    buf[3] = abs(dx) & 0xff;
    buf[4] = abs(dy) >> 8;
    buf[5] = abs(dy) & 0xff;
    Write(buf, 6);
    GetLine(resp);

    printf("translate (%d, %d)\n", dx, dy);


/*    int xc = 0, yc = 0;

    for(;;) {
        BYTE b[256];
        DWORD r;
        if(!ReadFile(Serial, b, sizeof(b), &r, NULL)) {
            printf("read error\n");
            exit(-1);
        }
        if(r == 0) break;
        for(int i = 0; i < r; i++) {
            if(b[i] == 'x') xc++;
            if(b[i] == 'y') yc++;
            if(b[i] == 'X') xc--;
            if(b[i] == 'Y') yc--;
            printf("%c", b[i]);
        }
    }

    printf("xc = %d yc = %d\n", xc, yc);
    if(xc != dx || yc != dy) {
        printf("error!\n");
        exit(-1);
    }
    printf("  okay\n"); */
}

void mill_now_drilling(void)
{
    int i;
    printf("raising router for tool change\n");
    for(i = 0; i < 20; i++) {
        mill_up();
    }
    printf("insert drill and position over pad0 (crosshairs in preview)\n");
    JogScreen();
    printf("now drilling\n");
}

void mill_nextsize(void)
{
}


void mill_drill(void)
{
    Slow();

    BYTE resp[10];
    BYTE cmd[1] = { DRILL };
    Write(cmd, 1);
    GetLine(resp);

    Fast();
}

void mill_down(void)
{
    Slow();

    BYTE resp[10];
    BYTE cmd[1] = { MILL_DOWN };
    Write(cmd, 1);
    GetLine(resp);
    // let it sink in!
    Sleep(100);

    Fast();
}

void mill_up(void)
{
    Slow();

    BYTE resp[10];
    BYTE cmd[1] = { MILL_UP };
    Write(cmd, 1);
    GetLine(resp);

    Fast();
}

//-----------------------------------------------------------------------------
// Enter a screen that allows the user to control each motor individually
// by hand.
//-----------------------------------------------------------------------------
static void JogScreen(void)
{
    printf("jog screen; try wsxWX edcEC rfvRV, qp when done\n"); 
    for(;;) {
        int c = _getch();
        switch(c) {
            case 'w': case 's': case 'x':
            case 'e': case 'd': case 'c':
            case 'r': case 'f': case 'v':
            case 'W': case 'X':
            case 'E': case 'C':
            case 'R': case 'V': {
                if(c == 'w') LastXWasNegative = TRUE;
                if(c == 'x') LastXWasNegative = FALSE;
                BYTE cmd[1] = { c };
                BYTE resp[5];
                Write(cmd, 1);
                GetLine(resp);
                break;
            }

            case 't':
                printf("+1000, +1000\n");
                mill_translate(1000, 1000);
                break;

            case 'o':
                printf("down\n");
                mill_down();
                break;

            case 'O':
                printf("up\n");
                mill_up();
                break;

            case 'q': {
                printf("type p to quit\n");
                int p = _getch();
                if(p != 'p') {
                    printf("never mind\n");
                } else {
                    return;
                }
                break;
            }

            default:
                printf("bad char\n");
                break;
        }
    }
}

//-----------------------------------------------------------------------------
// Try to get a ping from the controller..
//-----------------------------------------------------------------------------
static void GetPing(void)
{
    BYTE resp[10];
    BYTE cmd[1] = { PING_REQUEST };
    Write(cmd, 1);
    GetLine(resp);
    if(memcmp(resp, "pong", 4)==0) {
        printf("found controller\n");
    } else {
        printf("couldn't find controller\n");
        exit(-1);
    }
}

//-----------------------------------------------------------------------------
// Set speed, mill depth, drill depth.
//-----------------------------------------------------------------------------
static void InitParams(void)
{
    BYTE resp[10];

    BYTE setDrillDepth[2] = { SET_DRILL_STEPS, 200 };
    Write(setDrillDepth, 2);
    GetLine(resp);

    BYTE setMillDepth[2] = { SET_MILL_STEPS, 50 };
    Write(setMillDepth, 2);
    GetLine(resp);

    Fast();

    printf("configured controller\n");
}

//-----------------------------------------------------------------------------
// Open and configure the serial port, and put a handle to it in the global
// Serial. Print an error and exit if something fails.
//-----------------------------------------------------------------------------
static void InitSerial(void)
{
    Serial = CreateFile( "COM1", GENERIC_READ | GENERIC_WRITE, 0, NULL,
                     OPEN_EXISTING, 0, NULL);

    if (Serial == INVALID_HANDLE_VALUE) {
        printf("can't get port\n");
        exit(-1);
    }

    DCB dcb;

    if(!GetCommState(Serial, &dcb)) {
        printf("can't get params\n");
        exit(-1);
    }

    dcb.BaudRate = CBR_2400;
    dcb.ByteSize = 8;
    dcb.Parity   = NOPARITY;
    dcb.StopBits = ONESTOPBIT;

    if(!SetCommState(Serial, &dcb)) {
        printf("can't set params\n");
        exit(-1);
    }

    COMMTIMEOUTS cto;
    cto.ReadIntervalTimeout = 100;
    cto.ReadTotalTimeoutMultiplier = 0;
    cto.ReadTotalTimeoutConstant = 100;

    cto.WriteTotalTimeoutMultiplier = 0;
    cto.WriteTotalTimeoutConstant = 0;

    if(!SetCommTimeouts(Serial, &cto)) {
        printf("couldn't set timeouts\n");
        exit(-1);
    }
}

int main(int argc, char **argv)
{
    InitSerial();
    GetPing();
    InitParams();
    if(argc == 1) {
        JogScreen();
    } else {
        extern int mirror_x;
        char *file;
        if(argc == 3) {
            if(strcmp(argv[1], "--nomirror") != 0) {
                printf("usage: %s [ --nomirror ] [ file.scr ]\n", argv[0]);
                exit(-1);
            }
            mirror_x = 0;
            file = argv[2];
        } else {    
            file = argv[1];
        }
        if(mirror_x) {
            printf("position bit over front left of board\n");
        } else {
            printf("position bit over front right of board\n");
        }
        JogScreen();

        printf("will take an 0.010\" deep cut\n");

        mill_up();
        BYTE dropSet[2] = { SET_MILL_STEPS, 10 };
        Write(dropSet, 2);
        BYTE resp[5];
        GetLine(resp);

        mill_down();

        BYTE setMillDepth[2] = { SET_MILL_STEPS, 50 };
        Write(setMillDepth, 2);
        GetLine(resp);

        printf("now milling!\n");
        mill_translate(20, 20);
        mill_from_file(file);
    }
    return 0;
}
