#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <shlobj.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <shellapi.h>

#include "mill.h"

#define FREEZE_SUBKEY "WinRouter Preview"
#include "freeze.h"

#define TRANS_X(x) (((x) / Scale) + 20 + XOffset)
#define TRANS_Y(y) ((800 + YOffset - ((y) / Scale)))

HWND        MainWindow;
HDC         Hdc;
HFONT       Font;
HBRUSH      DrillBrush;
HPEN        DrillPen;
HBRUSH      MoveBrush;
HPEN        MovePen;
HBRUSH      OriginBrush;
HPEN        OriginPen;
int         DrillIndex;
BOOL        FirstDrill;
HBRUSH      DrillBrushes[4];
HPEN        DrillPens[4];
HINSTANCE   hInst;

int Scale = 1;

BOOL SlowMotion;

BOOL Drawing;
int  XPos;
int  YPos;

int  XRef;
int  YRef;
int  XLast;
int  YLast;

int  TotalDistance;
int  MilledDistance;

int XOffset;
int YOffset;

int XDrillOffset;
int YDrillOffset;

char InFile[MAX_PATH];

void dbp(char *str, ...)
{
    va_list f;
    char buf[1024];
    va_start(f, str);
    vsprintf(buf, str, f);
    OutputDebugString(buf);
    OutputDebugString("\n");
}

void mill_up(void)
{
    Drawing = FALSE;
}

void mill_down(void)
{
    Drawing = TRUE;
}

void mill_translate(int dx, int dy)
{
    if(Drawing) {
        SelectObject(Hdc, GetStockObject(WHITE_PEN));
        SelectObject(Hdc, GetStockObject(WHITE_BRUSH));
        MoveToEx(Hdc, TRANS_X(XPos), TRANS_Y(YPos), NULL);
        XPos += dx;
        YPos += dy;
        LineTo(Hdc, TRANS_X(XPos), TRANS_Y(YPos));
        if(SlowMotion) {
            MSG msg;
            Sleep(50);
            if(PeekMessage(&msg, NULL, 0, 0, TRUE)) {
                if(msg.message != WM_PAINT) {
                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }
            }
        }
    } else {
        SelectObject(Hdc, MovePen);
        SelectObject(Hdc, MoveBrush);
        MoveToEx(Hdc, TRANS_X(XPos), TRANS_Y(YPos), NULL);
        XPos += dx;
        YPos += dy;
        LineTo(Hdc, TRANS_X(XPos), TRANS_Y(YPos));
    }

    TotalDistance += (int)sqrt(dx*dx + dy*dy);
    if(Drawing) {
        MilledDistance += (int)sqrt(dx*dx + dy*dy);
    }
}

void mill_drill(void)
{
    int size = 8;
    SelectObject(Hdc, DrillBrush);
    SelectObject(Hdc, DrillPen);
    if(size < Scale)
        size = Scale;
    Ellipse(Hdc,
        TRANS_X(XPos - size), TRANS_Y(YPos - size),
        TRANS_X(XPos + size), TRANS_Y(YPos + size));

    if(FirstDrill) {
        int pSize = 30 / Scale;
        SelectObject(Hdc, OriginBrush);
        SelectObject(Hdc, OriginPen);
        MoveToEx(Hdc, TRANS_X(XPos) - pSize, TRANS_Y(YPos), NULL);
        LineTo  (Hdc, TRANS_X(XPos) + pSize + 1, TRANS_Y(YPos));
        MoveToEx(Hdc, TRANS_X(XPos), TRANS_Y(YPos) - pSize, NULL);
        LineTo  (Hdc, TRANS_X(XPos), TRANS_Y(YPos) + pSize + 1);
        FirstDrill = FALSE;
    }
}

void mill_nextsize(void)
{
    DrillBrush = DrillBrushes[DrillIndex];
    DrillPen   = DrillPens[DrillIndex];
    DrillIndex = (DrillIndex + 1) % 4;
}


void mill_now_drilling(void)
{
    FirstDrill = TRUE;
    XPos += XDrillOffset;
    YPos += YDrillOffset;
}

static void ShowOffset(int x, int y)
{
    char status[128];
    Hdc = GetDC(MainWindow);

    sprintf(status, "(dx, dy) = (%d, %d) thou                                ", 
        Scale * (x - XRef),
        Scale * (y - YRef));

    XLast = x;
    YLast = y;

    status[40] = '\0';

    SetBkColor(Hdc, RGB(0, 0, 0));
    SetTextColor(Hdc, RGB(255, 255, 255));
    SelectObject(Hdc, Font);

    TextOut(Hdc, 10, 70, status, strlen(status));

    ReleaseDC(MainWindow, Hdc);
}

LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
        case WM_PAINT: {
            extern int mirror_x;
            PAINTSTRUCT ps;
            char status[256];
            Hdc = BeginPaint(hwnd, &ps);

            XPos = 0;
            YPos = 0;

            SelectObject(Hdc, OriginBrush);
            SelectObject(Hdc, OriginPen);
            MoveToEx(Hdc, TRANS_X(XPos) - 5, TRANS_Y(YPos), NULL);
            LineTo  (Hdc, TRANS_X(XPos) + 6, TRANS_Y(YPos));
            MoveToEx(Hdc, TRANS_X(XPos), TRANS_Y(YPos) - 5, NULL);
            LineTo  (Hdc, TRANS_X(XPos), TRANS_Y(YPos) + 6);

            DrillIndex = 0;

            SetBkColor(Hdc, RGB(0, 0, 0));
            SetTextColor(Hdc, RGB(255, 255, 255));
            SelectObject(Hdc, Font);

            if(SlowMotion) {
                TextOut(Hdc, 10, 10, "[painting]", strlen("[painting]"));
            }

            TotalDistance = 0;
            MilledDistance = 0;

            mill_from_file(InFile);

            sprintf(status, "%d thou per pixel; drill offset (%d,%d); "
                "total displacement = %d.%03d\", milled %d.%03d\"; %s" , 
                Scale, XDrillOffset, YDrillOffset,
                TotalDistance / 1000, TotalDistance % 1000,
                MilledDistance / 1000, MilledDistance % 1000,
                mirror_x ? "mirrored" : "not mirrored");
            TextOut(Hdc, 10, 10, status, strlen(status));
            strcpy(status,
                "ASDW moves drill; +- zooms; <Up><Down><Left><Right> pans");
            TextOut(Hdc, 10, 30, status, strlen(status));
            strcpy(status,
                "<Shift>+... moves faster, Z homes drill, <Home> homes pan");
            TextOut(Hdc, 10, 50, status, strlen(status));
            SlowMotion = FALSE;

            ShowOffset(XLast, YLast);

            EndPaint(hwnd, &ps);
            return 1;
        }

        case WM_KEYDOWN: {
            volatile int s = GetAsyncKeyState(VK_SHIFT);
            volatile int shifted = GetAsyncKeyState(VK_SHIFT);
            int d;
            if(shifted) {
                d = 20;
            } else {
                d = 2;
            }
            switch(wParam) {
                case 'P':
                    SlowMotion = TRUE;
                    break;

                case 'Q':
                    PostQuitMessage(0);
                    break;

                case VK_SUBTRACT:
                case VK_OEM_MINUS:
                    Scale++;
                    break;

                case VK_ADD:
                case VK_OEM_PLUS:
                    if(Scale > 1)
                        Scale--;
                    break;

                case VK_LEFT:
                    XOffset += d*10;
                    break;

                case VK_RIGHT:
                    XOffset -= d*10;
                    break;

                case VK_UP:
                    YOffset += d*10;
                    break;

                case VK_DOWN:
                    YOffset -= d*10;
                    break;

                case 'A':
                    XDrillOffset -= d;
                    break;

                case 'D':
                    XDrillOffset += d;
                    break;

                case 'W':
                    YDrillOffset += d;
                    break;

                case 'S':
                    YDrillOffset -= d;
                    break;

                case 'Z':
                    XDrillOffset = 0;
                    YDrillOffset = 0;
                    break;

                case 'M': {
                    extern int mirror_x;
                    mirror_x = !mirror_x;
                    break;
                }
                    
                case VK_HOME:
                    XOffset = 0;
                    YOffset = 0;
                    break;
                
                case VK_SHIFT:
                    return 1;
            }
            InvalidateRect(hwnd, NULL, TRUE);
            return 1;
        }

        case WM_LBUTTONDOWN:
            XRef = LOWORD(lParam);
            YRef = HIWORD(lParam);
            ShowOffset(XRef, YRef);
            return 1;

        case WM_MOUSEMOVE:
            ShowOffset(LOWORD(lParam), HIWORD(lParam));
            return 1;

        case WM_CLOSE:
        case WM_DESTROY:
            PostQuitMessage(0);
            return 1;

        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }

}

static BOOL MakeWindowClass()
{
    WNDCLASSEX wc;
    memset(&wc, 0, sizeof(wc));
    wc.cbSize = sizeof(wc);

    wc.style            = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW | CS_OWNDC;
    wc.lpfnWndProc      = (WNDPROC)MainWndProc;
    wc.hInstance        = hInst;
    wc.hbrBackground    = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wc.lpszClassName    = "WinRouter";
    wc.lpszMenuName     = NULL;
    wc.hCursor          = LoadCursor(NULL, IDC_ARROW);

    return RegisterClassEx(&wc);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, INT nCmdShow)
{
    hInst = hInstance;

    strcpy(InFile, lpCmdLine);

    char buf[MAX_PATH+100];
    sprintf(buf, "WinRouter Preview [%s]", InFile);

    MakeWindowClass();

    Font = CreateFont(16, 0, 0, 0, FW_REGULAR, FALSE, FALSE, FALSE,
        ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
        FF_DONTCARE, "Tahoma");

    DrillBrushes[0] = CreateSolidBrush(RGB(255, 0, 0));
    DrillBrushes[1] = CreateSolidBrush(RGB(40, 255, 40));
    DrillBrushes[2] = CreateSolidBrush(RGB(50, 50, 255));
    DrillBrushes[3] = CreateSolidBrush(RGB(255, 255, 0));

    DrillPens[0] = CreatePen(PS_SOLID, 1, RGB(255, 0, 0));
    DrillPens[1] = CreatePen(PS_SOLID, 1, RGB(40, 255, 40));
    DrillPens[2] = CreatePen(PS_SOLID, 1, RGB(50, 50, 255));
    DrillPens[3] = CreatePen(PS_SOLID, 1, RGB(255, 255, 0));

    MovePen = CreatePen(PS_SOLID, 1, RGB(80, 80, 80));
    MoveBrush = CreateSolidBrush(RGB(80, 80, 80));

    OriginPen   = CreatePen(PS_SOLID, 1, RGB(255, 100, 10));
    OriginBrush = CreateSolidBrush(RGB(0, 0, 0));

    ThawDWORD(XOffset);
    ThawDWORD(YOffset);
    ThawDWORD(XDrillOffset);
    ThawDWORD(YDrillOffset);
    ThawDWORD(Scale);

    MainWindow = CreateWindowEx(0, "WinRouter", buf, WS_OVERLAPPED
        | WS_THICKFRAME | WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_SYSMENU,
        10, 10, 400, 400, NULL,
        NULL, hInst, NULL);

    ThawWindowPos(MainWindow);

    ShowWindow(MainWindow, SW_SHOW);

    MSG msg;
    DWORD ret;
    while(ret = GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    dbp("freezing");
    FreezeWindowPos(MainWindow);
    FreezeDWORD(XOffset);
    FreezeDWORD(YOffset);
    FreezeDWORD(XDrillOffset);
    FreezeDWORD(YDrillOffset);
    FreezeDWORD(Scale);

    return 0;
}
